#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"


extern int  VERBOSE;
extern long location_to_index(double, double, double);


//**************************************************************************

void ExpressionObj::Optimize()
{
  if (term_num == 0) return;

  for (int term=0; term < term_num; term++)
      if ( term_array[term].type == NUMBER_TYPE)
        {
        int tt = term;
        double v = term_array[tt].val;
        while ( (term_array[tt+1].type == NUMBER_TYPE) && (tt + 1 < term_num) )
          { v *= term_array[tt+1].val; tt++; }
        if (tt > term)
  	  {
          int dt = tt - term;
          term_array[term].val = v;
          term_num -= dt;
          for (tt = term + 1; tt < term_num; tt++)        
	    term_array[tt] = term_array[tt + dt];
	  }
	}
}

//**************************************************************************

inline double plus_minus(char op, double &a, double b)
{
  switch(op) {
   case T_PLUS: return a += b;
   case T_MINUS: return a -= b;
   case T_OR:  return a = ( (a > 0) || (b > 0) ? 1: 0 );
   case T_AND: return a = ( (a > 0) && (b > 0) ? 1: 0 );
   case T_GT:  return a = (a > b ? 1 : 0);
   case T_GE:  return a = (a >= b ? 1 : 0);
   case T_LT:  return a = (a < b ? 1 : 0);
   case T_LE:  return a = (a <= b ? 1 : 0);
   case T_EQ:  return a = (a == b ? 1 : 0);
   case T_NEQ: return a = (a != b ? 1 : 0);

   default: fprintf(stderr,"!!! Operation [%d=%c] not recognized\n", op, op);
            exit(1);
  }
}

//**************************************************************************

inline double mult_div(char op, double &a, double b)
{
  if  (op == T_MULT) return  a *= b;
    else if (op == T_DIV) return a /= b;
      else if (op == T_POWER) return a = pow(a,b);
        else if (op == T_MOD) return a = double ( int(a) % int(b) );
  else fprintf(stderr,"!!! Operation [%d=%c] not recognized\n", op, op);
  exit(1);
}

//**************************************************************************

double function(char op, double x)
{
  switch (op)
    {
    case BRACKET:   return x;
    case T_NOT:     return (x > 0) ? 0 : 1;
    case T_INT:     return double(int(x));
    case T_COSH:    return cosh(x);
    case T_SINH:    return sinh(x);
    case T_COS:     return cos(x);
    case T_SIN:     return sin(x);
    case T_TANH:    return tanh(x);
    case T_TAN:     return tan(x);
    case T_ATAN:    return atan(x);
    case T_EXP:     return exp(x);
    case T_LOG:     return log(x);
    case T_LOG10:   return log10(x);
    case T_SQR:     return x * x;
    case T_CUBE:    return x * x * x;
    case T_SQRT:    return sqrt(x);
    case T_THETA:   if (x == 0.0 ) return 0.5; else return (x > 0) ? 1 : 0;
    case T_SIGMA:   return 0.5 * (1 + tanh(x) );
    default:      fprintf(stderr,"function: Function [%c] not recognized\n", op);
                  exit(1);
    }
}

//**************************************************************************

double ExpressionObj::Evaluate()
  {
  if (term_num <= 0 ) return 0.0;

  const    int LEVELS = 5;
  double   result[LEVELS], value[LEVELS];
  unsigned char  plusmin[LEVELS], multdiv[LEVELS], func[LEVELS];
  bool     first[LEVELS];
  char     level;
  bool     Minus = false;
  double   val;

  struct TermStruct term;

  level = 0;
  result[0] = 0.0;    value[0] = 0.0;
  plusmin[0] = T_PLUS;  multdiv[0] = T_MULT;
  func[0] = BRACKET;  first[0] = true;

  for (int j=0; j < term_num; j++)

    switch ( (term = term_array[j]).type )
      {
      case T_PLUS: case T_MINUS:       // catch unary sign operators here
	                  if (j > 0) {
                            if ( term_array[j-1].type <= TYPE_NUM && term_array[j-1].type != BR_CLOSE) {
                              if ( term.type == T_MINUS ) Minus = true;
                              break;
			    }
			  }
			  else {
                            if ( term.type == T_MINUS ) Minus = true;
                            break;
			  }
      case T_AND: case T_OR: case T_GT: case T_GE: case T_LT: case T_LE: case T_EQ: case T_NEQ:   
                          plus_minus(plusmin[level], result[level], value[level]);
	                  value[level] = 1; plusmin[level] = term.type; 
                          break;

      case T_POWER:
      case T_MOD:
      case T_DIV:         multdiv[level] = term.type; break;
      case T_MULT:        break;

      case BR_CLOSE:      result[level] = 
                          function(func[level], plus_minus(plusmin[level], result[level], value[level]) );
	                  level --;
                          if (first[level]) { value[level] = result[level + 1]; first[level] = false; }
                            else mult_div(multdiv[level], value[level], result[level + 1] );
                          multdiv[level] = T_MULT;
                          break;

      case NUMBER_TYPE:   val = ( Minus ? -term.val : term.val ); Minus = false;
                          if (first[level]) { value[level] = val; first[level] = false; }
                            else mult_div(multdiv[level], value[level], val );
                          multdiv[level] = T_MULT;
                          break;

      case POINTER_TYPE:  val = ( Minus ? -*term.ptr : *term.ptr ); Minus = false;
                          if (first[level]) { value[level] = val; first[level] = false; }
                            else mult_div(multdiv[level], value[level], val );
                          multdiv[level] = T_MULT;
                          break;

      default:          // opening bracket or function
         	if (term.type >= TYPE_NUM) 
		  {
		  fprintf(stderr,"*** ExpressionObj::Evaluate: function \"%d\" not recognized (term %d of %d)\n",
                          term.type,j,term_num);
                  exit(1);
		  }

	        if ( (++level) > LEVELS ) {
                  fprintf(stderr, "*** ERROR: bracket nesting limit exceeded in expression { ");
                  print(); fprintf(stderr," }\n");
                  exit(1);
		}
                result[level] = 0.0;      value[level] = 0.0;
                plusmin[level] = T_PLUS;    multdiv[level] = T_MULT;
                func[level] = term.type;  first[level] = true;  break;
      }
  
  if (level != 0) { 
    fprintf(stderr, "*** Bracket nesting error in expression { ");
    print(); fprintf(stderr," }\n");
    exit(1);
  }
               
  return  function(func[0], plus_minus(plusmin[0], result[0], value[0]) );
  }

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//***************************************************************************

bool isFloat(const char *s)
{
 bool pointFlag = false, expFlag = false;
 int  last = strlen(s) - 1;
 int  i;

 for (i=0; i < strlen(s); i++) {
   if ( is_numeral(s[i]) ) continue;
   else if ( s[i] == '.' ) { 
     if (pointFlag) return false; 
     else pointFlag = true; 
   } 
   else if ( is_sign(s[i]) ) {
     if ( i == last ) return false;
     if ( (i == 0) && !( is_numeral(s[1]) || (s[1] == '.') ) ) return false;
     if ( (i > 0) && ( !( s[i-1] == 'e' || s[i-1] == 'E' ) || !is_numeral(s[i+1]) ) ) return false;
   }
   else if ( s[i] == 'e' || s[i] == 'E' ) {
     if ( (i == last) || (i == 0) ) return false;
     if ( !is_numeral(s[i-1]) || !(is_numeral(s[i+1]) || is_sign(s[i+1])) ) return false;
     if (expFlag) return false;
     expFlag = true;
   }
   else return false;
 }
 return true;
}

//***************************************************************************

bool isInt(const char *s)
{
 int  i;

 for (i=0; i < strlen(s); i++) {
   if ( is_numeral(s[i]) ) continue;
   else if ( is_sign(s[i]) ) { if (i > 0 ) return false; }
   else return false;
 }
 return true;
}

//**************************************************************************

KinExpression::KinExpression(TokenString &Param, int p0, double *arg, const char *argid, 
                             FieldObj *Ca, BufferArray *Bufs, KineticObj *KO, int *pLast) :
  ExpressionObj( get_max_term_num(Param, p0) )
{
  int tp;
  int p = p0;
  int term, term_num0 = term_num;

  for (term = 0; term < term_num; term++, p++) {
     if (VERBOSE > 7)  fprintf(stderr,"expression: term = %s \n", Param[p] );

    for (tp = 0; tp < TYPE_NUM; tp++)
      if (Param.equal(p, type_id[tp])) { 
	if (VERBOSE > 7)  fprintf(stderr,"it's a term = %s\n",type_id[tp]);
        term_array[term].type = tp; 

        if ( isBinaryOnly(tp) )    // catch a binary operator as first token, or two binaries in a row
          if (term == 0)  term_num = term; 
          else if ( !isOperand(term_array[term-1].type) && term_array[term-1].type != BR_CLOSE ) term_num = term;

        if ( term == 0 && tp == BR_CLOSE ) term_num = term;  // catch closing bracket as first token
        break;
      }

    if (tp < TYPE_NUM) { if (term_num < term_num0) break; else continue; }

    if ( Param.equal(p,argid) && arg) {
      term_array[term].type = POINTER_TYPE;
      term_array[term].ptr = arg;
      term_array[term].val = 0;
    }

    else if (Param.isNumber( Param[p], &(term_array[term].val)) ) { 
      if (VERBOSE > 7)  fprintf(stderr,"it's a number = %g\n", term_array[term].val );
      term_array[term].type = NUMBER_TYPE;
      term_array[term].ptr = 0;
    }

    else if ( term_array[term].ptr = ResolveID( Param[p], 0, Ca, Bufs, KO) ) {
      if (VERBOSE > 7)  fprintf(stderr,"it's a pointer = %ld\n", &term_array[term].ptr);
      term_array[term].type = POINTER_TYPE;
      term_array[term].val = 0;
    }

    else  { term_num = term; break; }  // does not match anything

  } // loop until end of line


  if (term_num) {                           // run some more checks on the last expression

    int last = term_array[term_num-1].type; // expression can only end in a number or a bracket    
    while ( !isOperand(last) && (last != BR_CLOSE ) && (term_num > 1) ) 
      last = term_array[--term_num-1].type;  

    if (term_num == 1)                     // one-term expression is either a pointer or a number
      if ( (last != NUMBER_TYPE ) && (last != POINTER_TYPE ) ) term_num = 0;
  } 

  if (pLast) *pLast = p0 + term_num - 1;
  else if ( term_num < term_num0 ) {
       fprintf(stderr, "\n*** KinExpression::KinExpression: invalid expression (token #%d):\n",p);
       Param.linePrint(p0 + term_num);
       exit(1);      
     }
   
  if (VERBOSE > 6) 
    { fprintf(stderr,"\nExpression = { "); print(stderr, Ca, Bufs, KO); fprintf(stderr," }"); }
  // Optimize(); 
}

//**************************************************************************

int get_max_term_num(TokenString &Param, int p)
{
  int terms = 0;

  while ( !isExpressionStop( Param[p] ) && ( p < Param.token_num ) ) 
     { p++; terms++; }

  return terms;
}

//**************************************************************************


void ExpressionObj::print(FILE *f, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO)
{
  char *s; 

  for (int term=0; term < term_num; term++)
   switch( term_array[term].type )
     {
     case NUMBER_TYPE:  fprintf(f,"%g ",term_array[term].val);
                        break;
     case POINTER_TYPE: if (s = ResolvePtr(term_array[term].ptr, Ca, Bufs, KO) )
                            fprintf(f,"%s ", s);
                        else fprintf(f,"t ");   
                          // correct above in the future: any way to tell if this pointer is really time?
			  // fprintf(f,"[%X] ", term_array[term].ptr );
                        break;
     default:           fprintf(f,"%s ", type_id[term_array[term].type]);
                        break;
     }
 fflush(f);
}


//**************************************************************************

double *ResolveID(char *svar, double **t, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO)
{
  int j;

  if (VERBOSE > 8) fprintf(stderr, ">> ResolveID: Resolving %s\n", svar);

   if (Ca)
     if ( equal(svar, Ca->ID) )  
       { if (t) *t = &Ca->Time; return (*Ca)(); }

    if (Bufs)
      for (j = 0; j < Bufs->buf_num; j++)
	if ( equal(svar,Bufs->array[j]->ID) )  
          { if (t) *t = &Bufs->array[0]->Time;  return (*Bufs->array[j])(); }

    if (KO)
      {
      if (t) *t = &KO->Time;

      if (KO->var) 
        for (j = 0; j < KO->var->get_size(); j++) 
          if (equal(svar, KO->var_id[j])) 
            return (*KO->var)() + j;

      if (KO->ident)
        for (j = 0; j < KO->ident->get_size(); j++)
          if (equal(svar, KO->ident_id[j])) 
            return  (*KO->ident)() + j;

      for (j = 0; j < KO->tables->table_num; j++)
	 if (equal(svar, KO->tables->array[j]->ID ) )
           return &(KO->tables->array[j]->value);

      for (j = 0; j < KO->maxima->peak_num; j++)
	 if (equal(svar, KO->maxima->array[j]->ID ) ) 
           return &(KO->maxima->array[j]->peak);

      for (j = 0; j < KO->locations->interpol_num; j++)
	 if (equal(svar, KO->locations->array[j]->ID ) ) 
	   {
           if (t) *t = &(KO->locations->array[j]->eigentime);
	   return &(KO->locations->array[j]->result);
	   }
      }

    //fprintf(stderr,"ResolveID: cannot interpret [%s]\n", svar ), exit(1);
    return 0;
}

//**************************************************************************


char *ResolvePtr(double *ptr, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO)
{
int j;
    if (Ca)  if ( ptr == Ca->elem )  return Ca->ID; 

    if (Bufs)
      for (j = 0; j < Bufs->buf_num; j++)
	if ( ptr == Bufs->array[j]->elem )  return  Bufs->array[j]->ID; 

    if (KO)
      {
      if (KO->var) 
        for (j = 0; j < KO->var->get_size(); j++)
         if ( ptr == ((*KO->var)() + j) ) return KO->var_id[j];
          
      if (KO->ident)
        for (j = 0; j < KO->ident->get_size(); j++)
         if ( ptr == ((*KO->ident)() + j) ) return KO->ident_id[j];

      for (j = 0; j < KO->tables->table_num; j++)
	 if ( ptr == &(KO->tables->array[j]->value) ) return KO->tables->array[j]->ID;

      for (j = 0; j < KO->maxima->peak_num; j++)
	 if ( ptr == &(KO->maxima->array[j]->peak) ) return KO->maxima->array[j]->ID;

      for (j = 0; j < KO->locations->interpol_num; j++)
	 if ( ptr == &(KO->locations->array[j]->result) ) return KO->locations->array[j]->ID;
      }

    return 0;
    //fprintf(stderr,"ResolvePtr: cannot resolve pointer %X\n", ptr );
    //exit(1);
}


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


