#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"


//***************************************************************************

TokenString::TokenString(const char *fname, const char *extra)
{
  FILE   *f;

  if (!(f = fopen(fname,"r"))) 
    { fprintf(stderr,"\n*** ERROR: TokenString::TokenString:: no file %s\n", fname); exit(1); }

  storage   = new char[max_length = MAX_STORAGE_LENGTH];
  token_ptr = new char *[max_token_num = MAX_TOKEN_NUM];

  length = 0;
  token_num = 0;
  token_ptr[0] = storage;

  parse(fname, extra);
  if (VERBOSE > 6) print();
}

//***************************************************************************

void TokenString::parse(const char *fname, const char *extra) {
  FILE   *f;
  bool exprFlag = false;   // in expression, minus/plus signs are treated as separate tokens

  int   ifStart[IF_NEST_LEVELS], ifLevel = -1;
  bool  ifClause[IF_NEST_LEVELS], waitForEndif[IF_NEST_LEVELS], waitForElse[IF_NEST_LEVELS];

  if (!(f = fopen(fname,"r"))) 
    { fprintf(stderr,"\n*** WARNING: TokenString::parse: no file \"%s\"\n", fname); return; }

  char lineString[MAX_LINE_LENGTH], token[MAX_TOKEN_LENGTH];

  while (1)   {
    if ( (token_num == 0) && extra ) strcpy(lineString, extra);
    else { fgets(lineString,MAX_LINE_LENGTH,f); if (feof(f)) break; }
    if ( strlen(lineString) >= MAX_LINE_LENGTH-1 )
      fprintf(stderr,"\n*** WARNING: max line length exceeded");

    char *ptr0 = lineString, *ptr1, *last = lineString + strlen(lineString);
    int tnum_old = token_num;
    
    do { 
      getToken(ptr0,ptr1,last, exprFlag);
      if ( (ptr1 - ptr0) >= MAX_TOKEN_LENGTH-1 )
        fprintf(stderr,"\n*** WARNING: max token length exceeded: token=%s", ptr0);
      strncpy(token,ptr0,ptr1-ptr0); token[ptr1-ptr0] = '\0';
      if ( ::equal(token, "debug") ) VERBOSE = 7;

      for (int i = 0; i < ALIAS_NUM; i++) 
	if (::equal(aliasArray[2*i], token) ) strcpy( token, aliasArray[2*i+1] );

      if ( ::equal(token, COMMENT_TOKEN) || ::equal(token, LINE_CONTINUE_TOKEN) )  break;

      if ( isExpressionStart(token) ) exprFlag = true;  // is an expression

      if ( ::equal(token, IF_TOKEN) ) {
         if ( ++ifLevel == IF_NEST_LEVELS ) 
            fprintf(stderr,"\n*** ERROR in TokenString::parse(): exceeded \"if\" nesting limit\n"), exit(1);
         ifStart[ifLevel] = token_num + 1;
         waitForEndif[ifLevel] = waitForElse[ifLevel] = false;
         if (ifLevel > 0 )
           if ( waitForEndif[ifLevel-1] || waitForElse[ifLevel-1] )  waitForEndif[ifLevel] = true;  // skip next if
      }

      if ( ifLevel >= 0 )
        if ( ( waitForEndif[ifLevel]  && !::equal(token, ENDIF_TOKEN) ) ||
	     ( waitForElse[ifLevel]   && !(::equal(token, ELSE_TOKEN) || ::equal(token, ENDIF_TOKEN) ) ) )  
             { ptr0 = ptr1; continue; }

      if ( ::equal(token, THEN_TOKEN) ) { 
        if ( ifLevel < 0 ) 
          fprintf(stderr,"\n>> TokenString::parse: [%s] without an [%s]\n",THEN_TOKEN,IF_TOKEN), exit(1);
	double condition = KinExpression(*this, ifStart[ifLevel]).Evaluate();
        ifClause[ifLevel] = ( condition > 0 ) ? true : false;
        if ( !ifClause[ifLevel] ) waitForElse[ifLevel] = true;
      }

      else if ( ::equal(token, ELSE_TOKEN) ) {
        if (ifClause[ifLevel]) waitForEndif[ifLevel] = true; else waitForElse[ifLevel] = false; 
        }

      else if ( ::equal(token, ENDIF_TOKEN) ) ifLevel--;

      else if ( ::equal(token,"include") ) {  // import data file
        if ( (ptr0 = ptr1) >= last ) break;
        getToken(ptr0,ptr1,last, exprFlag);
        char *input_file = ::StrCpy(ptr0,ptr1-ptr0);
        parse(input_file);
        delete [] input_file;
        if ( (ptr0 = ptr1) >= last ) break; 
          else continue;
      }
      else if ( ::equal(token,"echo") ) {  // echo a message
        if ( (ptr0 = ptr1) >= last ) break;
        getToken(ptr0,ptr1,last, exprFlag);
        strncpy(token,ptr0,ptr1-ptr0); token[ptr1-ptr0] = '\0';
        char *s = token_string(token);
        fprintf(stderr,"\n >>> USER MESSAGE: %s\n",s);
        delete [] s;
        if ( (ptr0 = ptr1) >= last ) break; 
          else continue;
      }

      addToken(token);
      ptr0 = ptr1;

    } while (ptr1 < last );

    if ( !::equal(token, LINE_CONTINUE_TOKEN) && token_num > tnum_old ) {
      addToken(CARR_RET_TOKEN);
      exprFlag = false;   // by default each new line is not an expression
    }
  }

  fclose(f);
}

//***************************************************************************

void TokenString::getToken(char *&ptr0, char *&ptr1, char *last, bool expression_flag)
{
  int i;
  ptr1 = ptr0;

  while ( is_delim( *ptr0 ) && ( ptr0 < last ) )  ptr1 = ++ptr0;
  if (ptr1 == last) return;

  if (*ptr0 == '\"' || *ptr0 == '\'' || *ptr0 == '`') {
    char quote = *ptr0;
    ptr1 = ++ptr0;
    while ( *(++ptr1) != quote ) 
      if (ptr1 >= last) { 
        fprintf(stderr,"\n*** ERROR: unterminated string in parameter file:\n %s\n",ptr0);
        exit(1);
      }
    *ptr1 = ' ';  // erase closing quote
    return;
  }

  for (i = 0; i < ALIAS_NUM; i++)      // *** 1. Look for alias tokens first
     if (::equal_(ptr0,aliasArray[2*i]) ) { ptr1 += strlen(aliasArray[2*i]); return; }

  if (expression_flag)                 // *** 2. Look for numerical expression tokens next
    for (i=0; i<TYPE_NUM; i++) 
      if ( ::equal_(ptr0, type_id[i]) ) { ptr1 += strlen(type_id[i]); return; }

  for (i = 0; i < SPECIAL_NUM; i++)    // *** 3. Then look for other special (non-delimited) tokens
     if (::equal_(ptr0,specialArray[i]) ) { ptr1 += strlen(specialArray[i]); return; }


  if ( is_short_token(*ptr1) && !is_sign(*ptr1) ) { ptr1++; return; }   // 4. *** Single-character token

  if ( is_sign(*ptr1) ) ptr1++;         // *** 5. A sign is a separate token only in an expression

  if ( is_numeral(*ptr1) || (*ptr1 == '.' && is_numeral(ptr1[1]) ) ) {    // getting a number
    while ( is_numeral(*ptr1) ) ptr1++;
    if (*ptr1 == '.') ptr1++;
    while ( is_numeral(*ptr1) ) ptr1++;
    if ( (*ptr1 == 'e' || *ptr1 == 'E') && 
         ( is_numeral(ptr1[1]) || ( is_sign(ptr1[1]) && is_numeral(ptr1[2]) ) ) ) {  // getting exponent
      ptr1++;
      if ( is_sign(*ptr1) ) ptr1++;
      while ( is_numeral(*ptr1) ) ptr1++;
    }
    else                                                          // otherwise it's a weird identifier  
      while (!is_short_token(*ptr1) && !is_delim(*ptr1)) ptr1 ++; // ("plot 2D", for instance)
    return;
  }

  while (!is_short_token(*ptr1) && !is_delim(*ptr1))  // some sort of an identifier
    ptr1 ++;

  return;
}

//***************************************************************************

void TokenString::addToken(const char *token)
{
  int tlength = strlen(token);
  if (tlength == 0) return;

  if (length + tlength + 1 >= max_length)
    {
    char *buffer = new char[max_length += STORAGE_INCREASE];   
    for (long l = 0; l < length; l++) buffer[l] = storage[l];
    for (int i = 0; i <= token_num; i++) token_ptr[i] = buffer + (token_ptr[i] - storage);
    delete [] storage;
    storage = buffer;
    }

  if (token_num + 1 >= max_token_num)
    {
    char **tptrs = new char *[max_token_num += TOKEN_NUM_INCREASE];   
    for (int i = 0; i <= token_num; i++) tptrs[i] = token_ptr[i];
    delete [] token_ptr;
    token_ptr = tptrs;
    }

  strncpy( token_ptr[token_num], token, tlength );
  *(token_ptr[token_num] + tlength) = 0;

  if (VERBOSE > 6) fprintf(stderr, "%s ", token_ptr[token_num]);

  token_ptr[token_num+1] = token_ptr[token_num] + tlength + 1;
  token_num++;
  length += tlength + 1;
}

//***************************************************************************

TokenString::TokenString(const TokenString &tstring)
{
  token_num = tstring.token_num;
  length    = tstring.length;

  storage = new char[length];
  token_ptr = new char *[token_num + 1];
 
}

//***************************************************************************

int TokenString::token_count(const char *token, int *p)
{
  int cnt = 0;
  for (int i=0; i<token_num; i++)  
    if ( equal(i, token) ) { cnt ++; if (p) *p = i; }
  return cnt;
}

//***************************************************************************

int TokenString::token2_count(const char *token1, const char *token2, int *p)
{
  int cnt = 0;
  for (int i = 1; i < token_num; i++)
    if ( equal(i-1, token1) && equal(i, token2) ) { cnt ++; if (p) *p = i; }
  return cnt;
}

//***************************************************************************

int TokenString::token3_count(const char *token1, const char *token2, const char *token3, int *p)
{
  int cnt = 0;
  for (int i = 2; i < token_num; i++)
   if ( equal(i-2, token1) && equal(i-1, token2) && equal(i, token3) ) { cnt ++; if (p) *p = i; }
  return cnt;
}
//***************************************************************************

int TokenString::token_count_(const char *token, int *p)
{
  int cnt = 0;
  for (int i=0; i<token_num; i++)  if ( equal_(i, token) ) { cnt ++; if (p) *p = i; }
  return cnt;
}

//***************************************************************************

void TokenString::print(const char *fname)
{
  FILE *f;

  if ( !(f = fopen(fname, "w")) ) 
    fprintf(stderr,"\nTokenString::print: could not open %s\n", fname);

  print(f);
}

//***************************************************************************

void TokenString::print(FILE *f)
{

  fprintf(f, "\n### Parameter file parsed: { ");
  for (int i = 0; i < token_num; i++)
   fprintf(f, "%s ", token_ptr[i]);
  fprintf(f, " }\n");
}

//***************************************************************************

void TokenString::linePrint(int p, FILE *f)
{
  fprintf(f, "\n");

  int p0 = p, i, j;

  while (p0 >= 0 && !equal(p0, CARR_RET_TOKEN) ) p0--;
  p0 ++;
 
  for (i = p0; i <= lastInLine(p); i++)
   fprintf(f, " %s", token_ptr[i]);
  fprintf(f, "\n ");

  for (i = p0; i <= lastInLine(p); i++) 
    for (j=0; j< strlen(token_ptr[i]) + 1; j++) 
      if (i < p) fprintf(f," ");
      else fprintf(f,"^");
  

  fprintf(f, "\n");
}
//***************************************************************************

void TokenString::trail_pars(const char *token, int m, ... )
{
char    type;
int    *ipar;
double *dpar;
long   *lpar;
char   *cpar;
char   *spar;

 int cnt = 0, ind = 0;

  do  {
    if ( equal(token, ind) ) cnt ++;
    if (cnt == m) break;
    }    
  while ( ++ind < token_num );

  if (cnt < m)  {
    fprintf(stderr,"\nTokenString::trail_pars: less than %d instances of %s\n",m,token);
    exit(1);
    }

  va_list p;
  va_start(p, m);

  while (1)
    {
    ind ++; 
    if ( equal(ind, CARR_RET_TOKEN) ) break;
    type = char( va_arg(p, int) );
    if (type == 'E' ) break;

    switch (type)
      {
      case 'd': dpar =  va_arg(p, double *);
 	        *dpar = index_double(ind); break;
      case 'i': ipar =  va_arg(p, int *);   
                *ipar = index_int(ind); break;
      case 'l': lpar =  va_arg(p, long *);
                *lpar = index_long(ind); break;
      case 'c': cpar =  va_arg(p, char *);   sscanf(token_ptr[ind],"%c",cpar); 
                break;
      case 's': spar =  va_arg(p, char *);   
                index_string(ind,spar); break;
      default:  va_arg(p, int); break;  // ignore argument 
      }
    }
  va_end(p);
}

//***************************************************************************

void TokenString::trail_pars(const char *token1, const char *token2, int m, ... )
{
char    type;
int    *ipar;
double *dpar;
long   *lpar;
char   *cpar;
char   *spar;

 char *token = ::StrCpy(token1, token2);
 int  cnt = 0, ind = 0;

  do  {
    if ( equal(token, ind) ) cnt ++;
    if (cnt == m) break;
    }    
  while ( ++ind < token_num );

  if (cnt < m)  {
    fprintf(stderr,"\nTokenString::trail_pars: less than %d instances of %s\n",m,token);
    exit(1);
    }
  delete [] token;

  va_list p;
  va_start(p, m);

  while (1)
    {
    ind ++; 
    if ( equal(ind, CARR_RET_TOKEN) ) break;
    type = char( va_arg(p, int) );
    if (type == 'E' ) break;

    switch (type)
      {
      case 'd': dpar =  va_arg(p, double *);
 	        *dpar = index_double(ind); break;
      case 'i': ipar =  va_arg(p, int *);   
                *ipar = index_int(ind);    break;
      case 'l': lpar =  va_arg(p, long *);
                *lpar = index_long(ind);   break;
      case 'c': cpar =  va_arg(p, char *);   sscanf(token_ptr[ind],"%c",cpar); 
                break;
      case 's': spar =  va_arg(p, char *);   
                index_string(ind,spar);    break;
      default:  va_arg(p, int); break;  // ignore argument 
      }
    }
  va_end(p);
}

//***************************************************************************

int TokenString::token_index(const char *token, int n) 
     { 
     int cnt = 0;
     for (int i = 0; i < token_num; i++)
       {
       if ( equal(i, token) ) cnt ++;
       if (cnt == n) return i;
       }
     fprintf(stderr,"\nTokenString::token_index: less than %d instances of [%s]\n",n,token),
     exit(1);
     }

int TokenString::token2_index(const char *token1, const char *token2, int n)
     { 
     int cnt = 0;
     for (int i = 1; i < token_num; i++)
       {
       if ( equal(i-1, token1) && equal(i, token2) ) cnt ++;
       if (cnt == n) return i;
       }
     fprintf(stderr,"\nTokenString::token2_index: less than %d instances of [%s %s]\n",n,token1,token2),
     exit(1);
     }

int TokenString::token3_index(const char *token1, const char *token2, const char *token3, int n)
     { 
     int cnt = 0;
     for (int i = 2; i < token_num; i++)
       {
       if ( equal(i-2, token1) && equal(i-1, token2) && equal(i, token3) ) cnt ++;
       if (cnt == n) return i;
       }
     fprintf(stderr,"\nTokenString::token3_index: less than %d instances of [%s %s %s]\n",n,token1,token2,token3),
     exit(1);
     }

int TokenString::token_index_(const char *token, int n) 
     { 
     int cnt = 0;
     for (int i = 0; i < token_num; i++)
       {
       if ( equal_(i, token) ) cnt ++;
       if (cnt == n) return i;
       }
     fprintf(stderr,"\nTokenString::token_index: less than %d instances of [%s]\n",n,token),
     exit(1);
     }


//***************************************************************************

bool TokenString::isNumber(const char *s, double *result, double *t, const char *tstr,
                             class FieldObj *Ca, class BufferArray *Bufs, class KineticObj *KO) 
   { 
   if (::isFloat(s)) { if (result) sscanf(s, "%lf", result); return true; }
   else if ( !token2_count(s,ASSIGN_LABEL)) return false;

   int pLast;
   int pFirst = token2_index(s, ASSIGN_LABEL) + 1;

   double x = KinExpression( *this, pFirst, t, tstr, Ca, Bufs, KO, &pLast).Evaluate();
   if (result) *result = x;

   if ( pLast >= pFirst && isExpressionStop( token_ptr[pLast+1] ) ) return true;
   else return false;
   }

//***************************************************************************

double TokenString::token_double(const char *s)
   {
   double r;
   int n;

   if (::isFloat(s) ) sscanf(s,"%lf",&r);
   else if (n = token2_count(s,ASSIGN_LABEL))
       {
       int p = token2_index(s,ASSIGN_LABEL);
       r = KinExpression(*this, p+1).Evaluate();
       if (n > 1) 
         if ( n - token3_count("for",s,ASSIGN_LABEL) > 1 )
 	   fprintf(stderr,"\n*** WARNING: TokenString::token_double: more than one definition for [%s] \n",s);
       }
   else fprintf(stderr,"\nTokenString::token_double: [%s] not a float\n",s), exit(1);

   return r;
   }

//***************************************************************************

int TokenString::token_int(const char *s)
   {
   int r;
   int  n;

   if (::isInt(s) ) { sscanf(s,"%d",&r); return r; }

   if (n = token2_count(s,ASSIGN_LABEL))
       {
       if (n > 1) 
         if ( n - token3_count("for",s,ASSIGN_LABEL) > 1 )
	   fprintf(stderr,"\n** TokenString::token_int: more than one definition for [%s] ****\n",s);

       int p = token2_index(s,ASSIGN_LABEL);
       double x = KinExpression(*this, p+1).Evaluate();
    
       if ( (x - int(x)) < 1e-12 ) return int( x );

       fprintf(stderr,"\nTokenString::token_int: [%g] not an integer\n",x); 
       exit(1);
       }
   
   fprintf(stderr,"\nTokenString::token_int: [%s] not an integer\n",s);
   exit(1);
   }

//***************************************************************************

long TokenString::token_long(const char *s)
   {
   long r;
   int  n;

   if (::isInt(s) ) { sscanf(s,"%ld",&r); return r; }

   if (n = token2_count(s,ASSIGN_LABEL))
       {
       if (n > 1) 
         if ( n - token3_count("for",s,ASSIGN_LABEL) > 1 )
	   fprintf(stderr,"\n** TokenString::token_long: more than one definition for [%s] ****\n",s);

       int p = token2_index(s,ASSIGN_LABEL);
       double x = KinExpression(*this, p+1).Evaluate();
    
       if ( (x - long(x)) < 1e-12 ) return long( x );

       fprintf(stderr,"\nTokenString::token_long: [%g] not an integer\n",x); 
       exit(1);
       }
   
   fprintf(stderr,"\nTokenString::token_long: [%s] not an integer\n",s);
   exit(1);
   }

//***************************************************************************

char *TokenString::token_string(const char *sin, char *sout, 
                               class FieldObj *Ca, class BufferArray *Bufs, class KineticObj *KO)
   {
   int  n = token2_count(sin,ASSIGN_LABEL);
   strcpy(sout,"");

   if (n > 1) 
     fprintf(stderr,"\n** WARNING: TokenString::token_string: multiple definitions for [%s]\n",sin);

   if ( n == 0 )  strcpy(sout,sin);  // simple string
   else 
     {
     int p = token2_index(sin,ASSIGN_LABEL) + 1;
     line_string(p, sout, Ca, Bufs, KO);
     }
   return sout;
   }

//***************************************************************************

char *TokenString::line_string(int p, char *sout, class FieldObj *Ca, class BufferArray *Bufs, class KineticObj *KO)
{
  int pp, p1 = lastInLine(p);

  strcpy(sout,"");
  
  while (p <= p1) {

    double x = KinExpression(*this, p, 0, "arg", Ca, Bufs, KO, &pp).Evaluate(); 

    if (pp >= p) {
      char temp[40];
      p = pp + 1;
      sprintf(temp,"%g",x);
      strcat(sout,temp);
    }
    else {
      char *temp = index_string( p );
      strcat(sout, temp);
      p++;
      delete [] temp;
    }
  }

  return sout;
}

//***************************************************************************

char *TokenString::token_string(const char *sin, 
                                class FieldObj *Ca, class BufferArray *Bufs, class KineticObj *KO)
   {
     char sout[5000];
     token_string(sin, sout, Ca, Bufs, KO);
     return ::StrCpy(sout);
   }

//***************************************************************************

int TokenString::tokens_to_eol(int p)
{
  int    num = 0;

  while ( !equal(p, CARR_RET_TOKEN) && p < token_num ) { p++; num++; }

  return num;
}

//***************************************************************************

